@@ -11,7 +11,9 @@ from reasoner_validator.biolink import (
 from reasoner_validator.trapi import check_trapi_validity, TRAPISchemaValidator
 from reasoner_validator.trapi.mapping import MappingValidator, check_node_edge_mappings
 
-TEST_DATA_SAMPLE_SIZE = 10
+# Unspoken assumption here is that validation of results returned for
+# Biolink Model release compliance only needs to be superficial
+RESULT_TEST_DATA_SAMPLE_SIZE = 10
 
 
 class TRAPIResponseValidator(ValidationReporter):
@@ -47,67 +49,28 @@ class TRAPIResponseValidator(ValidationReporter):
             strict_validation=strict_validation
         )
 
-
-    @staticmethod
-    def sanitize_trapi_query(response: Dict):
+    def sanitize_trapi_query(self, response: Dict) -> Dict:
         """
 
         :param response: Dict full TRAPI Response JSON object
         :return: Dict, response with discretionary removal of content which
                        triggers (temporarily) unwarranted TRAPI validation failures
         """
-        # Typical TRAPI Response
-        # {
-        #     'status_code': 200,
-        #     'response_json': {
-        #         'message': {
-        #             'query_graph': {
-        #                 'nodes': {
-        #                     'a': {
-        #                         'ids': ['NCBITaxon:2860379'],
-        #                         'categories': ['biolink:OrganismTaxon'],
-        #                         'is_set': False,
-        #                         'constraints': []
-        #                     }, 'b': {
-        #                         'ids': None,
-        #                         'categories': ['biolink:OrganismTaxon'],
-        #                         'is_set': False,
-        #                         'constraints': []
-        #                     }
-        #                 },
-        #                 'edges': {
-        #                     'ab': {
-        #                         'subject': 'a',
-        #                         'object': 'b',
-        #                         'knowledge_type': None,
-        #                         'predicates': ['biolink:subclass_of'],
-        #                         'attribute_constraints': [],
-        #                         'qualifier_constraints': []
-        #                     }
-        #                 }
-        #             },
-        #             'knowledge_graph': {
-        #                 'nodes': {},
-        #                 'edges': {}
-        #             },
-        #             'results': []
-        #         },
-        #         'log_level': None,
-        #         'workflow': [
-        #             {'runner_parameters': None, 'id': 'lookup', 'parameters': None}
-        #         ]
-        #     }
-        # }
-        if 'workflow' in response:
-            workflow_steps: Dict = response['workflow']  # is a list of steps
+        if 'workflow' in response and response['workflow']:
+            # a 'workflow' is a list of steps, which are JSON object specifications
+            workflow_steps: List[Dict] = response['workflow']
             for step in workflow_steps:
                 if 'runner_parameters' in step and not step['runner_parameters']:
+                    self.report("warning.trapi.response.workflow.runner_parameters.null")
                     step.pop('runner_parameters')
                 if 'parameters' in step and not step['parameters']:
-                    step['parameters'] = dict()
+                    # There are some workflow types that have mandatory need for 'parameters'
+                    # but this should be caught in a later schema validation step
+                    self.report("warning.trapi.response.workflow.parameters.null")
+                    step.pop('parameters')
         return response
 
-    def check_compliance_of_trapi_response(self, response: Optional[Dict]):
+    def check_compliance_of_trapi_response(self, response: Optional[Dict], edges_limit: int = 100):
         """
         One stop validation of all components of a TRAPI-schema compliant
         Query.Response, including its Message against a designated Biolink Model release.
@@ -124,6 +87,9 @@ class TRAPIResponseValidator(ValidationReporter):
 
         :param response: Query.Response to be validated.
         :type response: Optional[Dict]
+        :param edges_limit: integer maximum number of edges to be validated in the knowledge graph. A value of zero
+                            triggers validation of all edges in the knowledge graph (could take awhile! Default: 100)
+        :type edges_limit: int
 
         :returns: Validator cataloging "information", "warning" and "error" messages (may be empty)
         :rtype: ValidationReporter
@@ -131,7 +97,7 @@ class TRAPIResponseValidator(ValidationReporter):
         if not (response and 'message' in response):
             self.report("error.trapi.response.empty")
 
-        response: Dict = self.sanitize_trapi_query(response)
+        response = self.sanitize_trapi_query(response)
 
         trapi_validator: TRAPISchemaValidator = check_trapi_validity(
             instance=response,
@@ -143,7 +109,7 @@ class TRAPIResponseValidator(ValidationReporter):
 
         status: Optional[str] = response['status'] if 'status' in response else None
         if status and status not in ["Success", "QueryNotTraversable", "KPsNotAvailable"]:
-            self.report("warning.trapi.response.status.unknown", status=status)
+            self.report("warning.trapi.response.status.unknown", identifier=status)
 
         message: Optional[Dict] = response['message']
         if not message:
@@ -152,7 +118,7 @@ class TRAPIResponseValidator(ValidationReporter):
         # Sequentially validate the Query Graph, Knowledge Graph then validate
         # the Results (which rely on the validity of the other two components)
         elif self.has_valid_query_graph(message) and \
-                self.has_valid_knowledge_graph(message):
+                self.has_valid_knowledge_graph(message, edges_limit):
             self.has_valid_results(message)
 
     @staticmethod
@@ -160,36 +126,65 @@ class TRAPIResponseValidator(ValidationReporter):
         """
 
         :param results: List, original list of Results
-        :return: List, TEST_DATA_SAMPLE_SIZE sized subset of Results
+        :return: List, RESULT_TEST_DATA_SAMPLE_SIZE sized subset of Results
         """
-        sample_size = min(TEST_DATA_SAMPLE_SIZE, len(results))
+        sample_size = min(RESULT_TEST_DATA_SAMPLE_SIZE, len(results))
         result_subsample = results[0:sample_size]
         return result_subsample
 
     @staticmethod
-    def sample_graph(graph: Dict) -> Dict:
+    def sample_graph(graph: Dict, edges_limit: int = 100) -> Dict:
         """
+        Only process a strict subsample of the TRAPI Response Message knowledge graph.
 
-        :param graph: Dict, original knowledge graph
-        :return: Dict, TEST_DATA_SAMPLE_SIZE sized subset of knowledge graph
+        :param graph: original knowledge graph
+        :type graph: Dict
+        :param edges_limit: integer maximum number of edges to be validated in the knowledge graph. A value of zero
+                            triggers validation of all edges in the knowledge graph (could take awhile! Default: 100)
+        :type edges_limit: int
+
+        :return: Dict, 'edges_limit' sized subset of knowledge graph
         """
-        kg_sample: Dict = {
-            "nodes": dict(),
-            "edges": dict()
-        }
-        sample_size = min(TEST_DATA_SAMPLE_SIZE, len(graph["edges"]))
-        n = 0
-        for key, edge in graph['edges'].items():
-            kg_sample['edges'][key] = edge
-            if 'subject' in edge and edge['subject'] in graph['nodes']:
-                kg_sample['nodes'][edge['subject']] = graph['nodes'][edge['subject']]
-            if 'object' in edge and edge['object'] in graph['nodes']:
-                kg_sample['nodes'][edge['object']] = graph['nodes'][edge['object']]
-            n += 1
-            if n > sample_size:
-                break
+        # We don't check for non-empty graphs here simply because the sole caller of this
+        # method is the 'has_valid_knowledge_graph' method, which filters out empty graphs
+        if edges_limit > 0:
+            kg_sample: Dict = {
+                "nodes": dict(),
+                "edges": dict()
+            }
+            # 'sample_size' will always be a positive number here.
+            # The kg_sample size will always be the 'sample_size'
+            # or less, the latter situation arising if the number
+            # of graph edges is smaller or some subject or
+            # object id's are missing in the nodes list.
+            sample_size = min(edges_limit, len(graph["edges"]))
+            n = 0
+            for key, edge in graph['edges'].items():
 
-        return kg_sample
+                kg_sample['edges'][key] = edge
+
+                if 'subject' in edge and \
+                        edge['subject'] in graph['nodes'] and \
+                        edge['subject'] not in kg_sample['nodes']:
+                    kg_sample['nodes'][edge['subject']] = graph['nodes'][edge['subject']]
+
+                if 'object' in edge and \
+                        edge['object'] in graph['nodes'] and \
+                        edge['object'] not in kg_sample['nodes']:
+                    kg_sample['nodes'][edge['object']] = graph['nodes'][edge['object']]
+
+                n += 1
+                if n == sample_size:
+                    break
+
+            return kg_sample
+
+        else:
+            # No pruning... just return the contents of the entire knowledge graph
+            return {
+                "nodes": graph["nodes"],
+                "edges": graph["edges"]
+            }
 
     def has_valid_query_graph(self, message: Dict) -> bool:
         """
@@ -231,12 +226,22 @@ class TRAPIResponseValidator(ValidationReporter):
         # Only 'error' but not 'info' nor 'warning' messages invalidate the overall Message
         return False if self.has_errors() else True
 
-    def has_valid_knowledge_graph(self, message: Dict) -> bool:
+    def has_valid_knowledge_graph(self, message: Dict, edges_limit: int = 100) -> bool:
         """
         Validate a TRAPI Knowledge Graph.
+
         :param message: input message expected to contain the 'knowledge_graph'
+        :type message: Dict
+        :param edges_limit: integer maximum number of edges to be validated in the knowledge graph. A value of zero
+                            triggers validation of all edges in the knowledge graph (could take awhile! Default: 100)
+        :type edges_limit: int
+
         :return: bool, False, if validation errors
         """
+        # This integrity constraint may not really be necessary
+        # since negative numbers are functionally inequivalent to zero
+        assert edges_limit >= 0, "The 'edges_limit' must be zero or a positive integer!"
+
         # The Knowledge Graph should not be missing
         if 'knowledge_graph' not in message:
             self.report(code="error.trapi.response.knowledge_graph.missing")
@@ -245,8 +250,8 @@ class TRAPIResponseValidator(ValidationReporter):
             # The Knowledge Graph should also not generally be empty? Issue a warning
             if not (
                     knowledge_graph and len(knowledge_graph) > 0 and
-                    "nodes" in knowledge_graph and len(knowledge_graph["nodes"]) > 0 and
-                    "edges" in knowledge_graph and len(knowledge_graph["edges"]) > 0
+                    "nodes" in knowledge_graph and knowledge_graph["nodes"] and
+                    "edges" in knowledge_graph and knowledge_graph["edges"]
             ):
                 self.report(code="warning.response.knowledge_graph.empty")
                 # An empty knowledge graph (warning) does not generally invalidate
@@ -258,7 +263,7 @@ class TRAPIResponseValidator(ValidationReporter):
 
                 # ...then if not empty, validate a subgraph sample of the associated
                 # Knowledge Graph (since some TRAPI response kg's may be huge!)
-                kg_sample = self.sample_graph(knowledge_graph)
+                kg_sample = self.sample_graph(graph=knowledge_graph, edges_limit=edges_limit)
 
                 # Verify that the sample of the knowledge graph is TRAPI compliant
                 trapi_validator: TRAPISchemaValidator = check_trapi_validity(
@@ -271,14 +276,15 @@ class TRAPIResponseValidator(ValidationReporter):
 
                 # Verify that the sample of the knowledge graph is
                 # compliant to the currently applicable Biolink Model release
-                biolink_validator: BiolinkValidator = check_biolink_model_compliance_of_knowledge_graph(
-                    graph=kg_sample,
-                    biolink_version=self.biolink_version,
-                    sources=self.sources,
-                    # the ValidationReporter calling this function *might*
-                    # have an explicit strict_validation override (if not None)
-                    strict_validation=self.strict_validation
-                )
+                biolink_validator: BiolinkValidator = \
+                    check_biolink_model_compliance_of_knowledge_graph(
+                        graph=kg_sample,
+                        biolink_version=self.biolink_version,
+                        sources=self.sources,
+                        # the ValidationReporter calling this function *might*
+                        # have an explicit strict_validation override (if not None)
+                        strict_validation=self.strict_validation
+                    )
                 if biolink_validator.has_messages():
                     self.merge(biolink_validator)
 
@@ -343,7 +349,7 @@ class TRAPIResponseValidator(ValidationReporter):
                     #     if not any([alias == object_id for alias in output_aliases for object_id in object_ids]):
                     #         validator.report(
                     #             code=error.results.missing_bindings,
-                    #             input_id=case[output_element],
+                    #             identifier=case[output_element],
                     #             output_node_binding=output_node_binding
                     #         )
                     #         # data_dump=f"Resolved aliases:\n{','.join(output_aliases)}\n" +
