@@ -20,12 +20,6 @@ logger = logging.getLogger(__name__)
 
 pp = PrettyPrinter(indent=4)
 
-# TODO: is there a better way to ensure that a Biolink Model compliance test
-#       runs quickly enough if the knowledge graph is very large?
-#       Limiting nodes and edges viewed may miss deeply embedded errors(?)
-_MAX_TEST_NODES = 1000
-_MAX_TEST_EDGES = 100
-
 
 def _get_biolink_model_schema(biolink_version: Optional[str] = None) -> Optional[str]:
     # Get Biolink Model Schema
@@ -160,7 +154,7 @@ class BiolinkValidator(ValidationReporter):
             #       probably no longer relevant to the community?
             if 'categories' in slots:
                 if not isinstance(slots["categories"], List):
-                    self.report(code="error.knowledge_graph.node.categories.not_array", node_id=node_id)
+                    self.report(code="error.knowledge_graph.node.categories.not_array", identifier=node_id)
                 else:
                     categories = slots["categories"]
                     node_prefix_mapped: bool = False
@@ -178,12 +172,12 @@ class BiolinkValidator(ValidationReporter):
                     if not node_prefix_mapped:
                         self.report(
                             code="warning.knowledge_graph.node.unmapped_prefix",
-                            node_id=node_id, categories=str(categories)
+                            identifier=node_id, categories=str(categories)
                         )
             else:
                 self.report(
                     code="error.knowledge_graph.node.category.missing",
-                    context=self.graph_type.value, node_id=node_id
+                    context=self.graph_type.value, identifier=node_id
                 )
 
             # TODO: Do we need to (or can we) validate here, any other Knowledge Graph node fields? Perhaps not yet?
@@ -193,7 +187,7 @@ class BiolinkValidator(ValidationReporter):
             if "ids" in slots and slots["ids"]:
                 ids = slots["ids"]
                 if not isinstance(ids, List):
-                    self.report(code="error.query_graph.node.ids.not_array", node_id=node_id)
+                    self.report(code="error.query_graph.node.ids.not_array", identifier=node_id)
                     # we'll pretend that the ids were mistakenly
                     # just a scalar string, then continue validating
                     ids = [ids]
@@ -204,7 +198,7 @@ class BiolinkValidator(ValidationReporter):
                 categories = slots["categories"]
                 if categories:
                     if not isinstance(categories, List):
-                        self.report(code="error.query_graph.node.categories.not_array", node_id=node_id)
+                        self.report(code="error.query_graph.node.categories.not_array", identifier=node_id)
                     else:
                         id_prefix_mapped: Dict = {identifier: False for identifier in ids}
                         for category in categories:
@@ -226,7 +220,7 @@ class BiolinkValidator(ValidationReporter):
                         if unmapped_ids:
                             self.report(
                                 code="warning.query_graph.node.ids.unmapped_to_categories",
-                                node_id=node_id,
+                                identifier=node_id,
                                 unmapped_ids=str(unmapped_ids),
                                 categories=str(categories)
                             )
@@ -237,7 +231,7 @@ class BiolinkValidator(ValidationReporter):
             if 'is_set' in slots:
                 is_set = slots["is_set"]
                 if is_set and not isinstance(is_set, bool):
-                    self.report(code="error.query_graph.node.is_set.not_boolean", node_id=node_id)
+                    self.report(code="error.query_graph.node.is_set.not_boolean", identifier=node_id)
             # else:  # a missing key or null "is_set" value is permitted in QNodes but defaults to 'False'
 
             # constraints  # TODO: how do we validate node constraints?
@@ -257,24 +251,24 @@ class BiolinkValidator(ValidationReporter):
         """
         element: Optional[Element] = self.bmt.get_element(name)
         if not element:
-            self.report(code=f"error.{context}.unknown", name=name)
+            self.report(code=f"error.{context}.unknown", identifier=name)
             return None
         if element.deprecated:
-            self.report(code=f"warning.{context}.deprecated", name=name)
+            self.report(code=f"warning.{context}.deprecated", identifier=name)
             # return None - a deprecated term is not treated as a failure but just as a warning
         if element.abstract:
             if self.strict_validation:
-                self.report(code=f"error.{context}.abstract",  name=name)
+                self.report(code=f"error.{context}.abstract",  identifier=name)
                 return None
             else:
-                self.report(code=f"info.{context}.abstract", name=name)
+                self.report(code=f"info.{context}.abstract", identifier=name)
         elif self.bmt.is_mixin(name):
             # A mixin cannot be instantiated ...
             if self.strict_validation:
-                self.report(code=f"error.{context}.mixin", name=name)
+                self.report(code=f"error.{context}.mixin", identifier=name)
                 return None
             else:
-                self.report(code=f"info.{context}.mixin", name=name)
+                self.report(code=f"info.{context}.mixin", identifier=name)
         return element
 
     def validate_attributes(self, edge_id: str, edge: Dict):
@@ -284,11 +278,11 @@ class BiolinkValidator(ValidationReporter):
         :return: None (validation messages captured in the 'self' BiolinkValidator context)
         """
         if 'attributes' not in edge:
-            self.report(code="error.knowledge_graph.edge.attribute.missing", edge_id=edge_id)
+            self.report(code="error.knowledge_graph.edge.attribute.missing", identifier=edge_id)
         elif not edge['attributes']:
-            self.report(code="error.knowledge_graph.edge.attribute.empty", edge_id=edge_id)
+            self.report(code="error.knowledge_graph.edge.attribute.empty", identifier=edge_id)
         elif not isinstance(edge['attributes'], List):
-            self.report(code="error.knowledge_graph.edge.attribute.not_array", edge_id=edge_id)
+            self.report(code="error.knowledge_graph.edge.attribute.not_array", identifier=edge_id)
         else:
             attributes = edge['attributes']
 
@@ -311,19 +305,33 @@ class BiolinkValidator(ValidationReporter):
             # Expecting ARA and KP 'aggregator_knowledge_source' attributes?
             found_ara_knowledge_source = False
             found_kp_knowledge_source = False
-            found_primary_or_original_knowledge_source = False
+
+            # also track primary_knowledge_source attribute cardinality now
+            found_primary_knowledge_source: List[str] = list()
 
             for attribute in attributes:
 
                 # Validate attribute_type_id
                 if 'attribute_type_id' not in attribute:
-                    self.report(code="error.knowledge_graph.edge.attribute.type_id.missing")
+                    self.report(
+                        code="error.knowledge_graph.edge.attribute.type_id.missing",
+                        identifier=str(edge_id)
+                    )
                 elif not attribute['attribute_type_id']:
-                    self.report(code="error.knowledge_graph.edge.attribute.type_id.empty")
+                    self.report(
+                        code="error.knowledge_graph.edge.attribute.type_id.empty",
+                        identifier=str(edge_id)
+                    )
                 elif 'value' not in attribute:
-                    self.report(code="error.knowledge_graph.edge.attribute.value.missing")
+                    self.report(
+                        code="error.knowledge_graph.edge.attribute.value.missing",
+                        identifier=str(edge_id)
+                    )
                 elif not attribute['value']:
-                    self.report(code="error.knowledge_graph.edge.attribute.value.empty")
+                    self.report(
+                        code="error.knowledge_graph.edge.attribute.value.empty",
+                        identifier=str(edge_id)
+                    )
                 else:
                     attribute_type_id: str = attribute['attribute_type_id']
                     value = attribute['value']
@@ -338,6 +346,7 @@ class BiolinkValidator(ValidationReporter):
                     if not is_curie(attribute_type_id):
                         self.report(
                             code="error.knowledge_graph.edge.attribute.type_id.not_curie",
+                            identifier=str(edge_id),
                             attribute_type_id=str(attribute_type_id)
                         )
                     else:
@@ -345,13 +354,14 @@ class BiolinkValidator(ValidationReporter):
                         prefix = attribute_type_id.split(":", 1)[0]
                         if prefix == 'biolink':
                             biolink_class = self.validate_element_status(
-                                context="knowledge_graph.attribute.type_id",
+                                context="knowledge_graph.edge.attribute.type_id",
                                 name=attribute_type_id
                             )
                             if biolink_class:
                                 if not self.bmt.is_association_slot(attribute_type_id):
                                     self.report(
                                         code="warning.knowledge_graph.edge.attribute.type_id.not_association_slot",
+                                        identifier=str(edge_id),
                                         attribute_type_id=str(attribute_type_id)
                                     )
 
@@ -373,21 +383,18 @@ class BiolinkValidator(ValidationReporter):
                                         # TODO: not interested in any other attribute_type_id's at this moment
                                         continue
 
-                                    if attribute_type_id in \
-                                            [
-                                                "biolink:primary_knowledge_source",
-                                                "biolink:original_knowledge_source"
-                                            ]:
-                                        found_primary_or_original_knowledge_source = True
-
                                     # ... now, check the infores values against various expectations
                                     for infores in value:
                                         if not infores.startswith("infores:"):
                                             self.report(
                                                 code="error.knowledge_graph.edge.provenance.infores.missing",
-                                                infores=infores
+                                                identifier=str(edge_id),
+                                                infores=str(infores)
                                             )
                                         else:
+                                            if attribute_type_id == "biolink:primary_knowledge_source":
+                                                found_primary_knowledge_source.append(infores)
+
                                             if ara_source and \
                                                     attribute_type_id == "biolink:aggregator_knowledge_source" and \
                                                     infores == ara_source:
@@ -403,35 +410,51 @@ class BiolinkValidator(ValidationReporter):
                         elif not self.bmt.get_element_by_prefix(prefix):
                             self.report(
                                 code="warning.knowledge_graph.edge.attribute.type_id.unknown_prefix",
+                                identifier=str(edge_id),
                                 attribute_type_id=str(attribute_type_id)
                             )
 
                         else:
                             self.report(
-                                code="info.attribute.type_id.non_biolink_prefix",
+                                code="info.knowledge_graph.edge.attribute.type_id.non_biolink_prefix",
+                                identifier=str(edge_id),
                                 attribute_type_id=str(attribute_type_id)
                             )
 
             # TODO: After all the attributes have been scanned,
             #       check for provenance. Treat as warnings for now.
             if ara_source and not found_ara_knowledge_source:
-                self.report(code="warning.knowledge_graph.edge.provenance.ara.missing")
+                self.report(
+                    code="warning.knowledge_graph.edge.provenance.ara.missing",
+                    identifier=str(edge_id),
+                    ara_source=ara_source
+                )
 
             if kp_source and not found_kp_knowledge_source:
                 self.report(
                     code="warning.knowledge_graph.edge.provenance.kp.missing",
+                    identifier=str(edge_id),
                     kp_source=kp_source,
                     kp_source_type=kp_source_type
                 )
 
-            if not found_primary_or_original_knowledge_source:
-                self.report(code="warning.knowledge_graph.edge.provenance.missing_primary")
+            if not found_primary_knowledge_source:
+                self.report(
+                    code="error.knowledge_graph.edge.provenance.missing_primary",
+                    identifier=str(edge_id)
+                )
+            elif len(found_primary_knowledge_source) > 1:
+                self.report(
+                    code="warning.knowledge_graph.edge.provenance.multiple_primary",
+                    identifier=str(edge_id),
+                    sources=",".join(found_primary_knowledge_source)
+                )
 
     def validate_attribute_constraints(self, edge_id: str, edge: Dict):
         if 'attribute_constraints' not in edge or edge['attribute_constraints'] is None:
             return  # nullable: true... missing key or None value is ok?
         elif not isinstance(edge['attribute_constraints'], List):
-            self.report(code="error.query_graph.edge.attribute_constraints.not_array", edge_id=edge_id)
+            self.report(code="error.query_graph.edge.attribute_constraints.not_array", identifier=edge_id)
         elif not edge['attribute_constraints']:
             return  # nullable: true... an empty 'attribute_constraints' array is ok?
         else:
@@ -443,14 +466,20 @@ class BiolinkValidator(ValidationReporter):
             qualifier_type_id: str = qualifier['qualifier_type_id']
             qualifier_value: str = qualifier['qualifier_value']
             try:
-                if not self.bmt.validate_qualifier(
+                if not self.bmt.is_qualifier(name=qualifier_type_id):
+                    self.report(
+                        code=f"error.{context}.qualifier.type_id.unknown",
+                        identifier=edge_id,
+                        qualifier_type_id=qualifier_type_id
+                    )
+                elif not self.bmt.validate_qualifier(
                         # TODO: temporary workaround, parse 'qualifier_type_id' to core name
                         qualifier_type_id=parse_name(qualifier_type_id),
                         qualifier_value=qualifier_value
                 ):
                     self.report(
-                        code=f"error.{context}.qualifier.invalid",
-                        edge_id=edge_id,
+                        code=f"error.{context}.qualifier.value.unresolved",
+                        identifier=edge_id,
                         qualifier_type_id=qualifier_type_id,
                         qualifier_value=qualifier_value
                     )
@@ -460,7 +489,7 @@ class BiolinkValidator(ValidationReporter):
                 logger.error(f"BMT validate_qualifier Exception: {str(e)}")
                 self.report(
                     code=f"error.{context}.qualifier.invalid",
-                    edge_id=edge_id,
+                    identifier=edge_id,
                     qualifier_type_id=qualifier_type_id,
                     qualifier_value=qualifier_value
                 )
@@ -471,7 +500,7 @@ class BiolinkValidator(ValidationReporter):
         if 'qualifiers' not in edge or edge['qualifiers'] is None:
             return  # nullable: true... missing key or None value is ok?
         elif not isinstance(edge['qualifiers'], List):
-            self.report(code="error.knowledge_graph.edge.qualifiers.not_array", edge_id=edge_id)
+            self.report(code="error.knowledge_graph.edge.qualifiers.not_array", identifier=edge_id)
         elif not edge['qualifiers']:
             return  # nullable: true... an empty 'qualifiers' array is ok?
         else:
@@ -499,7 +528,7 @@ class BiolinkValidator(ValidationReporter):
                 if not qualifier_set_entry['qualifier_set']:
                     self.report(
                         code="error.query_graph.edge.qualifier_constraints.qualifier_set.empty",
-                        edge_id=edge_id
+                        identifier=edge_id
                     )
                 else:
                     # We have a putative non-empty 'qualifier_set'
@@ -530,15 +559,14 @@ class BiolinkValidator(ValidationReporter):
             if not self.bmt.is_predicate(predicate):
                 self.report(
                     code=f"error.{context}.invalid",
-                    edge_id=edge_id,
+                    identifier=edge_id,
                     predicate=predicate
                 )
             elif self.minimum_required_biolink_version("2.2.0") and \
                     not self.bmt.is_translator_canonical_predicate(predicate):
                 self.report(
                     code=f"warning.{context}.non_canonical",
-                    context=self.graph_type.value,
-                    edge_id=edge_id,
+                    identifier=edge_id,
                     predicate=predicate
                 )
 
@@ -570,7 +598,7 @@ class BiolinkValidator(ValidationReporter):
 
         # Validate Subject node
         if not subject_id:
-            self.report(code=f"error.{context}.edge.subject.missing", edge_id=edge_id)
+            self.report(code=f"error.{context}.edge.subject.missing", identifier=edge_id)
         elif subject_id not in self.nodes:
             self.report(code=f"error.{context}.edge.subject.missing_from_nodes", object_id=subject_id)
 
@@ -579,7 +607,8 @@ class BiolinkValidator(ValidationReporter):
             if not predicate:
                 self.report(
                     code="error.knowledge_graph.edge.predicate.missing",
-                    context=self.graph_type.value, edge_id=edge_id
+                    context=self.graph_type.value,
+                    identifier=edge_id
                 )
             else:
                 self.validate_predicate(edge_id=edge_id, predicate=predicate)
@@ -588,9 +617,9 @@ class BiolinkValidator(ValidationReporter):
                 # Query Graphs can have a missing or null predicates slot
                 pass
             elif not isinstance(predicates, List):
-                self.report(code="error.query_graph.edge.predicate.not_array", edge_id=edge_id)
+                self.report(code="error.query_graph.edge.predicate.not_array", identifier=edge_id)
             elif len(predicates) == 0:
-                self.report(code="error.query_graph.edge.predicate.empty_array", edge_id=edge_id)
+                self.report(code="error.query_graph.edge.predicate.empty_array", identifier=edge_id)
             else:
                 # Should now be a non-empty list of CURIES which are valid Biolink Predicates
                 for predicate in predicates:
@@ -600,7 +629,7 @@ class BiolinkValidator(ValidationReporter):
 
         # Validate Object Node
         if not object_id:
-            self.report(code=f"error.{context}.edge.object.missing", edge_id=edge_id)
+            self.report(code=f"error.{context}.edge.object.missing", identifier=edge_id)
         elif object_id not in self.nodes:
             self.report(code=f"error.{context}.edge.object.missing_from_nodes", object_id=object_id)
 
@@ -635,10 +664,10 @@ class BiolinkValidator(ValidationReporter):
                     name=category
             )
             if biolink_class and not self.bmt.is_category(category):
-                self.report(code=f"error.{context}.node.category.unknown", node_id=node_id, category=category)
+                self.report(code=f"error.{context}.node.category.unknown", identifier=node_id, category=category)
                 biolink_class = None
         else:
-            self.report(code=f"error.{context}.node.category.missing", node_id=node_id)
+            self.report(code=f"error.{context}.node.category.missing", identifier=node_id)
 
         return biolink_class
 
@@ -655,12 +684,12 @@ class BiolinkValidator(ValidationReporter):
                     self.report(
                         code="warning.input_edge.node.id.unmapped_to_category",
                         context=context,
-                        node_id=node_id,
+                        identifier=node_id,
                         category=category
                     )
             # else, we will have already reported an error in validate_category()
         else:
-            self.report(code="error.input_edge.node.id.missing", context=context)
+            self.report(code="error.input_edge.node.id.missing", identifier=context)
 
     def check_biolink_model_compliance_of_input_edge(self, edge: Dict[str, str]):
         """
@@ -696,7 +725,7 @@ class BiolinkValidator(ValidationReporter):
         if not predicate:
             self.report(
                 code="error.input_edge.predicate.missing",
-                edge_id=edge_id
+                identifier=edge_id
             )
         else:
             self.validate_predicate(edge_id=edge_id, predicate=predicate)
@@ -716,7 +745,7 @@ class BiolinkValidator(ValidationReporter):
         :type graph: Dict
         """
         if not graph:
-            self.report(code="warning.graph.empty", context=self.graph_type.value)
+            self.report(code="warning.graph.empty", identifier=self.graph_type.value)
             return  # nothing really more to do here!
 
         # Access graph data fields to be validated
@@ -739,32 +768,18 @@ class BiolinkValidator(ValidationReporter):
             # else:  Query Graphs can omit the 'edges' tag
             edges = None
 
-        # I only do a sampling of node and edge content. This ensures that
-        # the tests are performant but may miss errors deeper inside the graph?
-        nodes_seen = 0
         if nodes:
             for node_id, details in nodes.items():
-
                 self.validate_graph_node(node_id, details)
 
-                nodes_seen += 1
-                if nodes_seen >= _MAX_TEST_NODES:
-                    break
-
             # Needed for the subsequent edge validation
             self.set_nodes(set(nodes.keys()))
 
-            edges_seen = 0
             if edges:
                 for edge in edges.values():
-
                     # print(f"{str(edge)}", flush=True)
                     self.validate_graph_edge(edge)
 
-                    edges_seen += 1
-                    if edges_seen >= _MAX_TEST_EDGES:
-                        break
-
 
 def check_biolink_model_compliance_of_input_edge(
         edge: Dict[str, str],
