@@ -80,8 +80,7 @@ class ValidationReporter:
         #             { <parameters of second reported message...> },
         #             etc...
         #         ],
-        #         # codes without parameters can just be set to None?
-        #         # TODO: fix if this doesn't work for JSON serialization
+        #         # codes without parameters can just be set to an empty list?
         #         "info.compliant.message": []
         #
         #     },
@@ -97,7 +96,20 @@ class ValidationReporter:
             str,  # message type (info/warning/error)
             Dict[
                 str,  # message 'code' as indexing key
-                List[Dict[str, str]]  # List of parameters (Maybe empty, if code doesn't have any associated parameters)
+                # Dictionary of 'identifier' indexed messages with parameters
+                # (Maybe None, if code doesn't have any additional parameters)
+                Optional[
+                    Dict[
+                        str,  # key is the message-unique template 'identifier' value of parameterized messages
+                        Optional[
+                            List[
+                                # Each reported message adds a dictionary of such parameters
+                                # to the list here; these are not guaranteed to be unique
+                                Dict[str, str]
+                            ]
+                        ]
+                    ]
+                ]
             ]
         ] = {
             "information": dict(),
@@ -201,15 +213,44 @@ class ValidationReporter:
         # Sanity check: that the given code has been registered in the codes.yaml file
         assert CodeDictionary.get_code_entry(code) is not None, f"ValidationReporter.report: unknown code '{code}'"
 
-        message_type = self.get_message_type(code)
-        message_set = self._message_type_name[message_type]
-        if code not in self.messages[message_set]:
-            self.messages[message_set][code] = list()
-        # TODO: how can **message content duplication be avoided here(?)
+        message_type_id = self.get_message_type(code)
+        message_type = self._message_type_name[message_type_id]
+        if code not in self.messages[message_type]:
+            self.messages[message_type][code] = None
         if message:
-            self.messages[message_set][code].append(message)
+            # Should have at least an "identifier" parameter
+            if self.messages[message_type][code] is None:
+                self.messages[message_type][code] = dict()
 
-    def add_messages(self, new_messages: Dict[str, Dict[str, List[Dict[str, str]]]]):
+            # If a message has any parameters, then one of them is
+            # expected to be a message indexing identifier
+            if "identifier" in message:
+                message_identifier = message.pop("identifier")
+                if not message:
+                    # the message_identifier was the only parameter to keep track of...
+                    self.messages[message_type][code][message_identifier] = None
+                else:
+                    # keep track of additional parameters in a list of dictionaries
+                    # (may have additional, currently unavoidable, content duplication?)
+                    if message_identifier not in self.messages[message_type][code] or \
+                            self.messages[message_type][code][message_identifier] is None:
+                        self.messages[message_type][code][message_identifier] = list()
+
+                    self.messages[message_type][code][message_identifier].append(message)
+
+        # else: additional parameters are None
+
+    def add_messages(self, new_messages: Dict[
+            str,  # message type (info/warning/error)
+            Dict[
+                str,  # message 'code' as indexing key
+
+                # List of Dictionaries of parameters
+                # (Maybe None, if specific code doesn't
+                # have additional associated parameters)
+                Optional[Dict[str, Optional[List[Dict[str, str]]]]]
+            ]
+    ]):
         """
         Batch addition of a dictionary of messages to a ValidationReporter instance.
         :param new_messages: Dict[str, Dict], with key one of "information", "warnings" or "errors",
@@ -218,35 +259,50 @@ class ValidationReporter:
         for message_type in self.messages:   # 'info', 'warning', 'error'
             if message_type in new_messages:
                 message_type_contents = new_messages[message_type]
-                for code in message_type_contents:   # codes.yaml message codes
+                for code, content in message_type_contents.items():   # codes.yaml message codes
                     if code not in self.messages[message_type]:
-                        self.messages[message_type][code] = list()
+                        self.messages[message_type][code] = None
+                    if content:
+                        # content is of type Dict[str, Optional[List[Dict[str, str]]]]
+                        # where dictionary keys are a set of message discriminating 'identifier'
+                        identifier: str
+                        parameters: Optional[List[Dict[str, str]]]
+                        for identifier, parameters in content.items():
+                            if self.messages[message_type][code] is None:
+                                self.messages[message_type][code] = dict()
+                            if parameters:
+                                # additional parameters seen?
+                                if identifier not in self.messages[message_type][code] or \
+                                        self.messages[message_type][code][identifier] is None:
+                                    self.messages[message_type][code][identifier] = list()
 
-                    # TODO: how can **message content duplication be avoided here(?)
-                    self.messages[message_type][code].extend(message_type_contents[code])
+                                self.messages[message_type][code][identifier].extend(parameters)
+                            else:
+                                # the message 'identifier' is the only parameter
+                                self.messages[message_type][code][identifier] = None
 
-    def get_messages(self) -> Dict[str, Dict[str, Optional[List[Dict[str, str]]]]]:
+    def get_messages(self) -> Dict[str, Dict[str, Optional[Dict[str, Optional[List[Dict[str, str]]]]]]]:
         """
         Get copy of all messages as a Python data structure.
         :return: Dict (copy) of all validation messages in the ValidationReporter.
         """
         return copy.deepcopy(self.messages)
 
-    def get_info(self) -> Dict[str, Optional[List[Dict[str, str]]]]:
+    def get_info(self) -> Dict[str, Optional[Dict[str, Optional[List[Dict[str, str]]]]]]:
         """
         Get copy of all recorded information messages.
         :return: List, copy of all information messages.
         """
         return copy.deepcopy(self.messages["information"])
 
-    def get_warnings(self) -> Dict[str, Optional[List[Dict[str, str]]]]:
+    def get_warnings(self) -> Dict[str, Optional[Dict[str, Optional[List[Dict[str, str]]]]]]:
         """
         Get copy of all recorded warning messages.
         :return: List, copy of all warning messages.
         """
         return copy.deepcopy(self.messages["warnings"])
 
-    def get_errors(self) -> Dict[str, Optional[List[Dict[str, str]]]]:
+    def get_errors(self) -> Dict[str, Optional[Dict[str, Optional[List[Dict[str, str]]]]]]:
         """
         Get copy of all recorded error messages.
         :return: List, copy of all error messages.
@@ -324,7 +380,7 @@ class ValidationReporter:
         #             "warnings": [
         #                 {
         #                     "warning.predicate.non_canonical": [
-        #                         {"predicate": "biolink:participates_in"}
+        #                         {"predicate": "biolink:participates_in"}  xxx deprecated? better check this one!
         #                     ]
         #                 }
         #             ],
@@ -346,21 +402,48 @@ class ValidationReporter:
         else:
             return False
 
-    def display(self, messages: Dict[str, List[Dict[str, str]]]) -> List[str]:
+    def display(
+            self,
+            messages: Dict[
+                str,  # unique validation message codes
+                Optional[
+                    Dict[
+                        str,  # template 'identifier' key value
+                        Optional[
+                            List[
+                                Dict[str, str]  # dictionary of other template parameters (if present)
+                            ]
+                        ]
+                    ]
+                ]
+            ]
+    ) -> List[str]:
         """
         This augmented message display wrapper prepends
         the Validation Reporter contextual prefix to one or more
         resolved coded validation messages.
 
-        :param messages: Dict[str,List[Dict[str,str]]], dictionary of messages where the keys are message codes,
-                         and the values are (possibly empty) lists of dictionary objects of message parameters.
+        :param messages: Dict[str,Optional[Dict[str, Optional[List[Dict[str,str]]]]]], dictionary of messages where
+                         the keys are validation message codes, and the values are a dictionary of messages subsets
+                         keyed by message template 'identifier' field values. If the template has no other parameters,
+                         the given key has a value of None; otherwise, a list of dictionaries is given that report
+                         the values of message-specific template parameters in addition to the 'identifier' parameter.
         :return: List[str], one or more resolved and contextualized Validation Reporter messages
         """
         # TODO: are missing messages an absolute error?
         assert len(messages) > 0
         decoded_messages: List[str] = list()
-        code: str
-        parameters: List[Dict[str, str]]
+        code: str  # code for specific validation message template
+        parameters: Optional[
+                Dict[
+                    str,  # message template 'identifier' key value
+                    Optional[
+                        List[
+                            Dict[str, str]  # dictionary of other template parameters (if present)
+                        ]
+                    ]
+                ]
+            ]
         for code, parameters in messages.items():
             decoded_messages.extend(
                 [self.prefix + message for message in CodeDictionary.display(code, parameters)]
