@@ -190,6 +190,135 @@ _TEST_RESULTS_2 = [
     ]
 
 
+# Sample edge 2
+_TEST_EDGES_4 = {
+       "edge_1": {
+           "subject": "NCBIGene:29974",
+           "predicate": "biolink:physically_interacts_with",
+           "object": "PUBCHEM.COMPOUND:597",
+           "attributes": [
+               {
+                   "attribute_type_id": "biolink:aggregator_knowledge_source",
+                   "value": "infores:molepro"
+               },
+               {
+                   "attribute_type_id": "biolink:primary_knowledge_source",
+                   "value": "infores:automat-text-mining-provider"
+               },
+               {
+                   "attribute_type_id": "biolink:primary_knowledge_source",
+                   "value": "infores:hmdb"
+               }
+           ]
+        }
+    }
+
+
+_TEST_KG_4 = {
+    'nodes': _TEST_NODES_1,
+    'edges': _TEST_EDGES_4
+}
+
+
+@pytest.mark.parametrize(
+    "query",
+    [
+        (   # Query 0 - No 'workflow' key
+            {},
+        ),
+        (   # Query 1 - Null 'workflow' key value
+            {
+                'workflow': None
+            },
+        ),
+        (  # Query 2 - Null 'workflow' key list
+            {
+                'workflow': []
+            },
+        ),
+        (  # Query 3 - 'runner_parameters' is Null
+            {
+                'workflow': [
+                    {
+                        'runner_parameters': None,
+                        'id': 'sort_results_score',
+                        'parameters': {"ascending_or_descending": "ascending"}
+                    }
+                ]
+            },
+        ),
+        (  # Query 4 - 'parameters' is Null
+            {
+                'workflow': [
+                    {'runner_parameters': {'allowlist': ["infores:aragorn"]}, 'id': 'lookup', 'parameters': None}
+                ]
+            },
+        ),
+        (  # Query 5 - both 'parameters' and 'runner_parameters' are Null
+            {
+                'workflow': [
+                    {'runner_parameters': None, 'id': 'lookup', 'parameters': None}
+                ]
+            },
+        )
+    ]
+)
+def test_sanitize_trapi_query(query: Tuple):
+    validator: TRAPIResponseValidator = TRAPIResponseValidator()
+    response: Dict = validator.sanitize_trapi_query(response=query[0])
+    # assert response == query[1]
+
+
+NUM_SAMPLE_NODES = 5
+
+SAMPLE_NODES: Dict = dict()
+for node_id in range(1, NUM_SAMPLE_NODES):
+    SAMPLE_NODES[f"n{node_id}"] = dict()
+
+SAMPLE_EDGES: Dict = dict()
+for subject_id in SAMPLE_NODES:
+    for object_id in SAMPLE_NODES:
+        if subject_id != object_id:
+            SAMPLE_EDGES[f"e{subject_id}{object_id}"] = {
+                "subject": subject_id,
+                "predicate": "biolink:related_to",
+                "object": object_id
+            }
+
+TEST_GRAPH: Dict = {
+    "nodes": SAMPLE_NODES,
+    "edges": SAMPLE_EDGES
+}
+
+
+@pytest.mark.parametrize(
+    "query",
+    [
+        (   # Query 0 - unlimited sample whole graph
+            0,  # edges_limit
+            len(SAMPLE_NODES),  # number of nodes returned
+            len(SAMPLE_EDGES)   # number of edges returned
+        ),
+        (   # Query 1 - sample just 2 edges
+            2,  # edges_limit
+            3,  # number of nodes returned
+            2   # number of edges returned
+        ),
+        (   # Query 2 - sample just edge_value is negative number (functionally equivalent to zero)
+            -1,  # edges_limit
+            len(SAMPLE_NODES),  # number of nodes returned
+            len(SAMPLE_EDGES)   # number of edges returned
+        )
+    ]
+)
+def test_sample_graph(query: Tuple[int, int, int]):
+    validator: TRAPIResponseValidator = TRAPIResponseValidator()
+    kg_sample: Dict = validator.sample_graph(graph=TEST_GRAPH, edges_limit=query[0])
+    assert kg_sample
+    assert len(kg_sample["nodes"]) == query[1]
+    assert len(kg_sample["edges"]) == query[2]
+
+
 @pytest.mark.parametrize(
     "query",
     [
@@ -464,7 +593,8 @@ _TEST_RESULTS_2 = [
             ""
         ),
         (
-            # Query 16 - Full Message, with strict validation and a non-null kp_source_type results in missing primary
+            # Query 16 - Full Message, with strict validation and
+            #            a non-null kp_source_type results in missing primary
             {
                 "message": {
                     "query_graph": _TEST_QG_1,
@@ -480,10 +610,30 @@ _TEST_RESULTS_2 = [
                 "kp_source_type": "aggregator"
             },
             True,
-            "warning.knowledge_graph.edge.provenance.missing_primary"
+            "error.knowledge_graph.edge.provenance.missing_primary"
         ),
         (
-            # Query 17 - Full Message, with non-strict validation
+            # Query 17 - Full Message, with strict validation and
+            #            non-null kp_source_type results, has multiple primary knowledge sources
+            {
+                "message": {
+                    "query_graph": _TEST_QG_1,
+                    "knowledge_graph": _TEST_KG_4,
+                    "results": _TEST_RESULTS_1
+                }
+            },
+            None,
+            None,
+            {
+                "ara_source": None,
+                "kp_source": "infores:molepro",
+                "kp_source_type": "aggregator"
+            },
+            True,
+            "warning.knowledge_graph.edge.provenance.multiple_primary"
+        ),
+        (
+            # Query 18 - Full Message, with non-strict validation
             {
                 "message": {
                     "query_graph": _TEST_QG_2,
@@ -498,7 +648,7 @@ _TEST_RESULTS_2 = [
             ""
         ),
         (
-            # Query 18 - Full Message, WITH strict validation - abstract category?
+            # Query 19 - Full Message, WITH strict validation - abstract category?
             {
                 "message": {
                     "query_graph": _TEST_QG_2,
@@ -513,7 +663,7 @@ _TEST_RESULTS_2 = [
             "error.query_graph.node.category.abstract"
         ),
         (
-            # Query 19 - Full Message, WITH strict validation - abstract predicate?
+            # Query 20 - Full Message, WITH strict validation - abstract predicate?
             {
                 "message": {
                     "query_graph": _TEST_QG_2,
@@ -528,7 +678,7 @@ _TEST_RESULTS_2 = [
             "error.query_graph.edge.predicate.abstract"
         ),
         (
-            # Query 20 - Valid full Message, under strict validation.
+            # Query 21 - Valid full Message, under strict validation.
             #            Message is valid, but the 'workflow' field is not an array?
             {
                 "message": {
@@ -546,7 +696,7 @@ _TEST_RESULTS_2 = [
             "error.trapi.validation"
         ),
         (
-            # Query 21 - Valid full Message, under strict validation.
+            # Query 22 - Valid full Message, under strict validation.
             #            Message is valid, the 'workflow' field is an array,
             #            but the single list entry is an invalid workflow spec?
             {
@@ -566,7 +716,7 @@ _TEST_RESULTS_2 = [
             "error.trapi.validation"
         ),
         (
-            # Query 22 - Valid full Message, under strict validation.
+            # Query 23 - Valid full Message, under strict validation.
             #            Message is valid, the 'workflow' field is an array,
             #            but the single list entry is in the workflow schema
             #            and has at least the one required field 'id'
@@ -585,7 +735,7 @@ _TEST_RESULTS_2 = [
             ""   # this simple workflow spec should pass?
         ),
         (
-            # Query 23 - Valid full Message, under strict validation. Message is valid, the 'workflow' field is array,
+            # Query 24 - Valid full Message, under strict validation. Message is valid, the 'workflow' field is array,
             #            but the single list entry is an elaborated 'real world' workflow spec,
             #            but one entry overlay_compute_ngd is incomplete - doesn't fully validate!
             {
@@ -629,7 +779,7 @@ _TEST_RESULTS_2 = [
             "error.trapi.validation"
         ),
         (
-            # Query 24 - Valid full Message, under strict validation. Message is valid, the 'workflow' field is array,
+            # Query 25 - Valid full Message, under strict validation. Message is valid, the 'workflow' field is array,
             #            but the single list entry is an elaborated 'real world' workflow spec
             {
                 "message": {
@@ -671,7 +821,7 @@ _TEST_RESULTS_2 = [
             ""   # this simple workflow spec should pass?
         ),
         (
-            # Query 25 - Valid full Message, under strict validation. Message is valid,
+            # Query 26 - Valid full Message, under strict validation. Message is valid,
             #            the 'workflow' field is an array, but runner_parameters is None.
             #            This is technically invalid but we have a code patch which should filter it out (for now)
             {
@@ -696,7 +846,7 @@ _TEST_RESULTS_2 = [
         )
     ]
 )
-def test_check_biolink_model_compliance_of_trapi_response(query: Tuple[Union[Dict, str]]):
+def test_check_biolink_model_compliance_of_trapi_response(query: Tuple):
     validator: TRAPIResponseValidator = TRAPIResponseValidator(
         trapi_version=query[1],
         biolink_version=query[2],
